from __future__ import annotations

import os
import time
from collections import deque
from enum import Enum
from typing import Any, List, Dict
from pathlib import PurePosixPath
import logging

from loki.common.profile import Profile, ProfileManager
from loki.scrapers.base import Method, RequestMixin

logger = logging.getLogger('loki')


class PathExpect(str, Enum):
    ANY = "any"
    DIR = "dir"
    FILE = "file"


class Pan115(RequestMixin):
    """
    不支持文件夹名称包含"/"
    """
    
    PLATFORM = "115"

    def __init__(
        self,
        *,
        profile: Profile | None = None,
        profile_name: str = "default",
        profile_manager: ProfileManager | None = None,
        auto_load_cookies: bool = True,
    ) -> None:
        if profile is None:
            manager = profile_manager or ProfileManager()
            profile = manager.get(self.PLATFORM, profile_name)

        super().__init__(profile=profile, auto_load_cookies=auto_load_cookies)
        config = profile.read_config()
        self.uid = str(config.get("uid", ""))
        if not self.uid:
            self.uid = self.get_uid()
            config['uid'] = self.uid
            profile.write_config(config=config)
        
    def get_uid(self) -> str:
        url = 'https://webapi.115.com/files'
        params = {
            'aid': '1',
            'cid': '0',
            'offset': '0',
            'limit': '20',
            'type': '0',
            'show_dir': '1',
            'fc_mix': '0',
            'natsort': '1',
            'count_folders': '1',
            'format': 'json',
            'custom_order': '0'
        }
        resp = self.request(Method.GET, url, params=params)
        if resp:
            data = resp.json()
            if data['state'] == True:
                return data['uid']
        return '' 
           
    def path_to_id(self, pan_path: str, expect: PathExpect = PathExpect.ANY) -> str:
        """
        开头有无/都视为绝对路径
        """     
        if not isinstance(expect, PathExpect):
            raise TypeError(f"expect must be PathExpect, got {type(expect).__name__}")
        id = '0'
        cur = PurePosixPath('/')
        p = '/' / PurePosixPath(pan_path)
        if len(p.parts) == 1:
            return id if expect in (PathExpect.ANY, PathExpect.DIR) else ""
        for part in p.parts[1:]:
            cur /= part
            is_last = (cur == p)
            children = self.list_files(id)
            dirs = [c for c in children if c.get('name') == cur.name and bool(c.get('is_dir', False))]
            files = [c for c in children if c.get('name') == cur.name and not bool(c.get('is_dir', False))]

            if not is_last:
                if not dirs:
                    return ''
                next_id = str(dirs[0].get('id', '') or '').strip()
                if not next_id:
                    return ''
                id = next_id
                continue

            if expect == PathExpect.DIR:
                chosen = dirs[0] if dirs else None
            elif expect == PathExpect.FILE:
                chosen = files[0] if files else None
            else:
                chosen = (dirs[0] if dirs else None) or (files[0] if files else None)
            if not chosen:
                return ""
            next_id = str(chosen.get('id', '') or '').strip()
            if not next_id:
                return ""
            id = next_id
        return id
    
    def paths_to_ids(
        self,
        paths: List[str],
        limit: str = '20',
        expect: PathExpect = PathExpect.ANY,
        sleep_seconds: float = 0.7,
    ) -> List[str]:
        """
        缓存遍历过的文件、文件夹，减少请求调用
        开头有无/都视为绝对路径
        不存在文件/文件夹 或 类型不匹配(expect) 的 id 为 -1
        """
        if not isinstance(expect, PathExpect):
            raise TypeError(f"expect must be PathExpect, got {type(expect).__name__}")
        path2meta_cache: Dict[str, Dict[str, str]] = {'/': {'dir': '0'}}
        dirid2files_cache: Dict[str, List[Dict[str, Any]]] = {}
        results = []
        
        for p_str in paths:
            p = '/' / PurePosixPath(p_str)
            p_posix = p.as_posix()
            cached = path2meta_cache.get(p_posix)
            if cached:
                if expect == PathExpect.DIR:
                    results.append(cached.get('dir', '') or '-1')
                elif expect == PathExpect.FILE:
                    results.append(cached.get('file', '') or '-1')
                else:
                    results.append(cached.get('dir', '') or cached.get('file', '') or '-1')
                continue
            
            cur = PurePosixPath('/')
            for part in p.parts[1:]:
                parent_meta = path2meta_cache.get(cur.as_posix())
                parent_dir_id = str((parent_meta or {}).get('dir', '') or '').strip()
                if not parent_dir_id:
                    break

                children = dirid2files_cache.get(parent_dir_id)
                if children is None:
                    if sleep_seconds > 0:
                        time.sleep(sleep_seconds)
                    children = self.list_files(parent_dir_id, limit)
                    dirid2files_cache[parent_dir_id] = children

                for child in children:
                    name = str(child.get('name', '') or '')
                    if not name:
                        continue
                    child_id = str(child.get('id', '') or '').strip()
                    if not child_id:
                        continue
                    child_path = (cur / name).as_posix()
                    meta = path2meta_cache.setdefault(child_path, {})
                    if bool(child.get('is_dir', False)):
                        meta['dir'] = child_id
                    else:
                        meta['file'] = child_id
                cur /= part
            resolved = path2meta_cache.get(p_posix)
            if not resolved:
                results.append('-1')
            elif expect == PathExpect.DIR:
                results.append(resolved.get('dir', '') or '-1')
            elif expect == PathExpect.FILE:
                results.append(resolved.get('file', '') or '-1')
            else:
                results.append(resolved.get('dir', '') or resolved.get('file', '') or '-1')
        
        return results
            
    def list_files(self, id: str = '0', limit: str = '20') -> List[Dict[str, Any]]:
        items: List[Dict[str, Any]] = []
        
        url = 'https://webapi.115.com/files'
        params = {
            'aid': '1',
            'cid': id,
            'offset': '0',
            'limit': limit,
            'type': '0',
            'show_dir': '1',
            'fc_mix': '0',
            'natsort': '1',
            'count_folders': '1',
            'format': 'json',
            'custom_order': '0'
        }
        resp = self.request(Method.GET, url, params=params)
        if resp:
            data = resp.json()
            if data['state'] == True and str(data['cid']) == id:
                for f in data['data']:
                    name = str(f.get('n', '') or '')
                    if not name:
                        continue

                    fid = str(f.get('fid', '') or '').strip()
                    if fid:
                        items.append({
                            'id': fid,
                            'name': name,
                            'size': int(f.get('s', 0)),
                            'class': str(f.get('class', '')),
                            'is_dir': False,
                        })
                        continue

                    # Folder name may contain suffix like "abc.mp4", so do not infer by extension.
                    cid = str(f.get('cid', '') or '').strip()
                    if not cid:
                        continue
                    items.append({
                        'id': cid,
                        'name': name,
                        'size': int(f.get('s', 0)),
                        'class': '',
                        'is_dir': True,
                    })
                
        return items

    def walk_path(self, pan_path: str, limit: str = '1000', sleep_seconds: float = 1) -> List[Dict[str, Any]]:
        """
        递归遍历指定目录下所有文件/文件夹。

        返回结果结构与 `list_files` 一致，并额外附带 `path` 字段（绝对路径）。

        注意：底层仍依赖 `list_files` 的 `limit`，目录项超过 limit 时结果可能不完整；
        且同一目录下如果存在同名“文件+文件夹”，会同时返回两条记录。
        """
        base = '/' / PurePosixPath(pan_path)
        base_id = self.path_to_id(base.as_posix(), expect=PathExpect.DIR)
        if not base_id:
            return []

        results: List[Dict[str, Any]] = []
        q = deque([(base, base_id)])
        first_request = True

        while q:
            cur_path, cur_id = q.popleft()
            if not first_request and sleep_seconds > 0:
                time.sleep(sleep_seconds)
            first_request = False
            children = self.list_files(cur_id, limit)
            for info in children:
                name = str(info.get('name', '') or '')
                if not name:
                    continue
                child_path = (cur_path / name).as_posix()
                entry = dict(info)
                entry['path'] = child_path
                results.append(entry)

                child_id = str(info.get('id', '') or '').strip()
                if child_id and bool(info.get('is_dir', False)):
                    q.append((PurePosixPath(child_path), child_id))

        return results
    
    def _offline_space(self) -> Dict:
        url = 'https://115.com/?ct=offline&ac=space'
        resp = self.request(Method.GET, url)
        if resp:
            return resp.json()
        return {}

    def add_lixian_task(self, links: List[str], pan_path: str) -> List[Dict]:
        """
        返回任务hash
        """
        if self.create_folder(pan_path) == '-1':
            return []
        
        wp_path_id = self.path_to_id(pan_path, expect=PathExpect.DIR)
        if not wp_path_id:
            return []

        return self.add_lixian_task_id(links, wp_path_id)
    
    def add_lixian_task_id(self, links: List[str], wp_path_id: str) -> List[Dict]:
        """
        返回任务hash（直接传入目录 id）
        """
        results = []
        wp_path_id = str(wp_path_id or '').strip()
        if not wp_path_id:
            return results
        off_space_data = self._offline_space()
        
        url = 'https://115.com/web/lixian/?ct=lixian&ac=add_task_urls'
        form_data = {
            'sign': off_space_data.get('sign', ''),
            'time': str(off_space_data.get('time', '')),
            'uid': self.uid,
            'savepath': '',
            'wp_path_id': wp_path_id,
        }
        for i, link in enumerate(links):
            form_data[f'url[{i}]'] = link
        
        resp = self.request(Method.POST, url, data=form_data)
        if resp:
            data = resp.json()
            if data.get('errno') == 0:
                tasks = data.get('result')
                for task in tasks:
                    results.append({
                        'info_hash': task.get('info_hash'),
                        'url': task.get('url'),
                        'name': task.get('name'),
                    })
        return results

    def get_lixian_tasks(self, type: int=12) -> Dict:
        """
        12: 正在下载
        11: 完成记录
        9: 下载失败
        
        结果说明:
        wp_path_id: 存储在的文件夹id
        file_id: 文件id(目录id)
        info_hash: 任务hash
        """
        off_space_data = self._offline_space()
        url = 'https://115.com/web/lixian/?ct=lixian&ac=task_lists'
        form_data = {
            'page': '1',
            'stat': str(type),
            'uid': self.uid,
            'sign': off_space_data.get('sign', ''),
            'time': str(off_space_data.get('time', '')),
        }
        resp = self.request(Method.POST, url, data=form_data)
        if resp:
            data = resp.json()
            if data['state'] == True:
                return data
        return {}
    
    def move(self, paths: List[str], dest_path: str, expect: PathExpect = PathExpect.ANY) -> bool:
        if not paths or not dest_path:
            return True
        
        if not isinstance(expect, PathExpect):
            raise TypeError(f"expect must be PathExpect, got {type(expect).__name__}")
        src_ids = self.paths_to_ids(paths, expect=expect)
        dest_id = self.create_folder(dest_path)
        if dest_id == '-1':
            return False
        return self.move_ids(src_ids, dest_id)

    def move_ids(self, src_ids: List[str], dest_id: str) -> bool:
        """
        直接传入 id 列表进行移动
        """
        if not src_ids:
            return True
        dest_id = str(dest_id or '').strip()
        if not dest_id or dest_id == '-1':
            return False
        url = 'https://webapi.115.com/files/move'
        
        form_data = {
            'pid': dest_id,
            'format': 'json',
        }
        j = 0
        for src_id in src_ids:
            src_id = str(src_id or '').strip()
            if not src_id or src_id == '-1':
                continue
            form_data[f'fid[{j}]'] = src_id
            j += 1
        if j == 0:
            return False
        
        resp = self.request(Method.POST, url, data=form_data)
        if resp:
            data = resp.json()
            if data['state'] == True:
                return True
        return False
    
    def delete(self, paths: List[str], expect: PathExpect = PathExpect.ANY) -> bool:
        if not paths:
            return True
        
        if not isinstance(expect, PathExpect):
            raise TypeError(f"expect must be PathExpect, got {type(expect).__name__}")
        ids_to_delete = self.paths_to_ids(paths, expect=expect)
        return self.delete_ids(ids_to_delete)

    def delete_ids(self, ids_to_delete: List[str]) -> bool:
        """
        直接传入 id 列表进行删除
        """
        if not ids_to_delete:
            return True
        url = 'https://webapi.115.com/rb/delete'
        
        form_data = {
            'ignore_warn': '0',
        }
        j = 0
        for src_id in ids_to_delete:
            src_id = str(src_id or '').strip()
            if not src_id or src_id == '-1':
                continue
            form_data[f'fid[{j}]'] = src_id
            j += 1
        if j == 0:
            return True
        
        resp = self.request(Method.POST, url, data=form_data)
        if resp:
            data = resp.json()
            if data['state'] == True:
                return True
        return False    
    
    def rename(self, paths2newname: Dict[str, str], expect: PathExpect = PathExpect.ANY) -> bool:
        paths = list(paths2newname.keys())
        if not isinstance(expect, PathExpect):
            raise TypeError(f"expect must be PathExpect, got {type(expect).__name__}")
        ids_to_rename = self.paths_to_ids(paths, expect=expect)
        
        id2newname: Dict[str, str] = {}
        for i in range(len(ids_to_rename)):
            rid = str(ids_to_rename[i] or '').strip()
            if not rid or rid == '-1':
                continue
            id2newname[rid] = paths2newname[paths[i]]
        return self.rename_ids(id2newname)

    def rename_ids(self, id2newname: Dict[str, str]) -> bool:
        """
        直接传入 id->新名字 进行重命名（不走 path_to_id/paths_to_ids）
        """
        if not id2newname:
            return False

        url = 'https://webapi.115.com/files/batch_rename'
        form_data = {
            'format': 'json'
        }
        j = 0
        for rid, newname in id2newname.items():
            rid = str(rid or '').strip()
            if not rid or rid == '-1':
                continue
            form_data[f'files_new_name[{rid}]'] = newname
            j += 1
        if j == 0:
            return False

        resp = self.request(Method.POST, url, data=form_data)
        if resp:
            data = resp.json()
            if data['state'] == True:
                return True
        return False
    
    def create_folder(self, folder_path: str) -> str:
        """
        返回创建的文件夹的id
        """
        return self.create_folder_to_id('0', folder_path)

    def create_folder_in(self, parent_id: str, folder_name: str) -> str:
        """
        在指定 parent_id 下创建一个子文件夹，返回新文件夹 id，失败返回 -1
        """
        parent_id = str(parent_id or '').strip()
        folder_name = str(folder_name or '').strip()
        if not parent_id or not folder_name:
            return '-1'

        url = 'https://webapi.115.com/files/add'
        form_data = {
            'pid': parent_id,
            'cname': folder_name
        }
        resp = self.request(Method.POST, url, data=form_data)
        if resp:
            data = resp.json()
            if data.get('state') == True:
                return str(data.get('cid', '') or '').strip() or '-1'
        return '-1'

    def create_folder_to_id(self, parent_id: str, folder_path: str, limit: str = '1000') -> str:
        """
        在指定 parent_id 下按 folder_path 创建多级子文件夹，返回最终文件夹 id，失败返回 -1。
        """
        parent_id = str(parent_id or '').strip()
        if not parent_id:
            return '-1'

        p = PurePosixPath(folder_path)
        parts = [seg for seg in p.parts if seg not in ('/', '')]
        if not parts:
            return parent_id

        cur_id = parent_id
        for seg in parts:
            children = self.list_files(cur_id, limit)
            existing_dir = next((c for c in children if c.get('name') == seg and bool(c.get('is_dir', False))), None)
            if existing_dir:
                next_id = str(existing_dir.get('id', '') or '').strip()
                if not next_id:
                    return '-1'
                cur_id = next_id
                continue

            new_id = self.create_folder_in(cur_id, seg)
            if new_id == '-1':
                return '-1'
            cur_id = new_id

        return cur_id

    def upload_file_id(self, local_path: str, pan_path_id: str) -> str:
        """
        不支持大文件分片上传
        """
        filename = PurePosixPath(local_path).name
        
         # 获取上传信息
        url = 'https://uplb.115.com/3.0/sampleinitupload.php'
        form_data = {
            'userid': self.uid,
            'filename': filename,
            'filesize': str(os.path.getsize(local_path)),
            'target': f'U_1_{pan_path_id}',
        }
        
        resp = self.request(Method.POST, url, data=form_data)
        if not resp:
            return '-1'
        # {"object":"69836aa2e4b80bc9fc4a9ecff879174b488390ad","accessid":"LTAI5tMvF188zm8v4bVG2BQC","host":"https:\/\/fhnfile.oss-cn-shenzhen.aliyuncs.com","policy":"eyJleHBpcmF0aW9uIjoiMjAyNi0wMi0wNFQyMzo1MDoyNFoiLCJjb25kaXRpb25zIjpbWyJjb250ZW50LWxlbmd0aC1yYW5nZSIsMCw1MzY4NzA5MTIwXV19","signature":"qw+QflqvMxMH4wDIJcwnCt4KrAU=","expire":1770220224,"callback":"eyJjYWxsYmFja1VybCI6Imh0dHA6XC9cL3VwbGIuMTE1LmNvbVwvMy4wXC9zYW1wbGVjb21wbGV0ZXVwbG9hZC5waHAiLCJjYWxsYmFja0JvZHkiOiJzaGExPSR7c2hhMX0mZmlsZW5hbWU9JHtvYmplY3R9JmZpbGVzaXplPSR7c2l6ZX0mbWltZVR5cGU9JHttaW1lVHlwZX0maGVpZ2h0PSR7aW1hZ2VJbmZvLmhlaWdodH0md2lkdGg9JHtpbWFnZUluZm8ud2lkdGh9JnVzZXJpZD0xMDMyMDE2NTQmdGFyZ2V0PVVfMV8zMzU3MDczMjM0NjcwMzg2ODQwJnVzZXJmbj1NRGhETUVFeE1UVTFSVVl6UlRKR1FrTTVOVVE1UWtJNE5URXpOamc1UlVSQ1F6QXpNRGxFTXk1emNuUT0mYnVja2V0PWZobmZpbGUmdXNlcmlwPTExMC43My44Ny41NyZ1c2VycG9ydD0zODAwMCZzb3VyY2U9NCZjYl90b2tlbj0zZTQyZTY3NDY5ODU4Yjg5YzJmYjBhM2E4NGNiNmU5ZCIsImNhbGxiYWNrQm9keVR5cGUiOiJhcHBsaWNhdGlvblwveC13d3ctZm9ybS11cmxlbmNvZGVkIn0="}
        upload_info = resp.json()
        upload_url = upload_info.get('host', '')
        if not upload_url:
            return '-1'
        files = {
            'file': open(local_path, 'rb')
        }
        form_data = {
            'name': filename,
            'policy': upload_info.get('policy', ''),
            'OSSAccessKeyId': upload_info.get('accessid', ''),
            'success_action_status': '200',
            'callback': upload_info.get('callback', ''),
            'signature': upload_info.get('signature', ''),
            'key': upload_info.get('object', ''),
            'target': f'U_1_{pan_path_id}',
        }
        resp = self.request(Method.POST, upload_url, data=form_data, files=files)

        if not resp:
            return '-1'
        data = resp.json()
        if data.get('state') == True:
            return data.get('data', {}).get('file_id', '-1')
        return '-1'

    def upload_files(self, local_paths: List[str], pan_path: str) -> List[str]:
        """
        不支持大文件分片上传
        """
        wp_path_id = self.create_folder(pan_path)
        if wp_path_id == '-1':
            return []
        results = []
        for local_path in local_paths:
            file_id = self.upload_file_id(local_path, wp_path_id)
            results.append(file_id)
            
        return results   
        


import os

import pytest

from loki.scrapers.javdb import JavDB


@pytest.mark.skipif(
    os.environ.get("LOKI_JAVDB_INTEGRATION") != "1",
    reason="integration test (requires network/playwright browser)",
)
def test_search_integration():
    jdb = JavDB()
    try:
        res = jdb.search("PRED-427")
        assert isinstance(res, dict)
    finally:
        jdb.close()



from __future__ import annotations

import logging
from playwright.sync_api import sync_playwright, Page

from loki.common.profile import Profile, ProfileManager
from loki.scrapers.base import RequestMixin

logger = logging.getLogger('loki')

class TikTok(RequestMixin):
    PLATFORM = "TikTok"
    
    def __init__(
        self,
        *,
        profile: Profile | None = None,
        profile_name: str = "default",
        profile_manager: ProfileManager | None = None,
        auto_load_cookies: bool = True,
    ) -> None:
        if profile is None:
            manager = profile_manager or ProfileManager()
            profile = manager.get(self.PLATFORM, profile_name)

        super().__init__(profile=profile, auto_load_cookies=auto_load_cookies)
        self._base = 'https://www.tiktok.com'
        self._playwright = sync_playwright().start()
        self._browser = self._playwright.chromium.launch(
            headless=False,
            args=[
                "--disable-blink-features=AutomationControlled",
            ]
        )
        
    def close(self) -> None:
        try:
            if getattr(self, "_browser", None):
                self._browser.close()
        finally:
            if getattr(self, "_playwright", None):
                self._playwright.stop()
                
    def _new_page(self) -> tuple[BrowserContext, Page] | tuple[None, None]:
        context = None
        try:
            context = self._browser.new_context(
                viewport={"width": 1280, "height": 800},
                user_agent="Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 Chrome/120 Safari/537.36",
                locale="zh-CN",
                timezone_id="Asia/Shanghai",
            )
            page = context.new_page()
            return context, page
        except Exception:
            logger.exception("创建页面失败")
            if context:
                context.close()
            return None, None
     
    def get_profile_data(self, username: str) -> dict | None:
        profile_url = f"{self._base}/@{username}"

        context, page = self._new_page()
        if not page:
            return None

        try:
            with page.expect_response(
                lambda r: ("api/post/item_list" in r.url)
                          and r.request.resource_type in ("xhr", "fetch")
                          and r.status == 200,
                timeout=30000
            ) as resp_info:
                page.goto(profile_url, wait_until="domcontentloaded", timeout=30000)

            data = resp_info.value.json()
            return data

        except Exception:
            logger.exception("抓取 profile 接口失败: %s", profile_url)
            return None
        finally:
            context.close()


from __future__ import annotations

import re
import time
from typing import List, Dict
import logging
from bs4 import BeautifulSoup, Tag
from playwright.sync_api import sync_playwright
from urllib.parse import urlencode

from loki.common.profile import Profile, ProfileManager
from loki.scrapers.base import Method, RequestMixin

logger = logging.getLogger('loki')

_CODE_CMP_RE = re.compile(r"([A-Z]{2,10})[-_ ]*(\d{2,6})", re.I)

class JavDB(RequestMixin):
    PLATFORM = "JavDB"

    def __init__(
        self,
        *,
        profile: Profile | None = None,
        profile_name: str = "default",
        profile_manager: ProfileManager | None = None,
        auto_load_cookies: bool = True,
    ) -> None:
        if profile is None:
            manager = profile_manager or ProfileManager()
            profile = manager.get(self.PLATFORM, profile_name)

        super().__init__(profile=profile, auto_load_cookies=auto_load_cookies)
        self._base = 'https://javdb.com'
        self._playwright = sync_playwright().start()
        self._browser = self._playwright.chromium.launch(
            headless=True,
            args=[
                "--disable-blink-features=AutomationControlled",
                "--start-maximized"
            ]
        )
        
    def _get_soup(self, url: str, params: Dict = {}) -> BeautifulSoup | None:
        try:
            context = self._browser.new_context(
                viewport={"width": 1280, "height": 800},
                user_agent="Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 Chrome/120 Safari/537.36",
                locale="zh-CN",
                timezone_id="Asia/Shanghai"
            )
            page = context.new_page()
            try:
                page.goto(f'{url}?{urlencode(params)}', wait_until="domcontentloaded", timeout=10000)
                return BeautifulSoup(page.content(), 'html.parser')
            finally:
                context.close()  # 关键：每次用完关 context
        except Exception:
            logger.exception("请求异常 %s", url)
        return None   
    
    def close(self):
        self._browser.close()
        self._playwright.stop()
        
    def _norm_code(self, s: str | None) -> str | None:
        if not s:
            return None
        s = str(s).upper().replace('.', '-').replace('_', '-').replace(' ', '-')
        m = _CODE_CMP_RE.search(s)
        if not m:
            return None
        prefix = m.group(1).upper()
        digits = (m.group(2) or '').lstrip('0') or '0'
        return f"{prefix}-{digits}"
    
    def _abs_url(self, href: str | None) -> str:
        h = (href or "").strip()
        if not h:
            return ""
        if h.startswith("http://") or h.startswith("https://"):
            return h
        if not h.startswith("/"):
            h = "/" + h
        return self._base + h
    
    def search(self, code: str, has_magnet: bool = True, no_subtitle: bool=True, no_uncensored: bool=True) -> Dict:
        search_code = self._norm_code(code) or code
        result = {
            'code': search_code
        }
        
        url = f'https://javdb.com/search'
        params = {
            'q': search_code,
            'f': 'all'
        }
        if has_magnet:
            params['f'] = 'download'
            
        soup = self._get_soup(url, params=params)
        if not soup:
            return result
        
        target_url = ''
        items = soup.select(".movie-list .item")
        for it in items:
            title_el = it.select_one("a .video-title strong")
            link_el = it.select_one("a[href]")
            if not title_el or not link_el:
                continue
            found_text = title_el.get_text(strip=True)
            found_norm = self._norm_code(found_text)
            # 只比较“前缀-数字”，忽略数字后的 A/B/CH 等
            if found_norm and found_norm == search_code:
                href = link_el.get("href", "")
                if isinstance(href, str):
                    target_url = self._abs_url(href.strip())
                    break
        
        if target_url:
            soup = self._get_soup(target_url)
            if not soup:
                return result
            
            actress_a = soup.select_one('.movie-panel-info a:has(+ .female)')
            if isinstance(actress_a, Tag):
                result['actress_name'] = actress_a.text
            
            items = soup.select("#magnets-content .item")
            best_magnet = ''
            best_size_gib = -1.0
        
            def size_from_meta(txt: str) -> float:
                # 如 "10.44GB, 1個文件" / "700MB, 2 files"
                s = txt.upper()
                m = re.search(r"([0-9]+(?:\.[0-9]+)?)\s*(GB|G|MB|M)", s)
                if not m:
                    return 0.0
                val = float(m.group(1))
                unit = m.group(2)
                if unit.startswith("G"):
                    return val
                return val / 1024.0
        
            for it in items:
                a = it.select_one("a[href^='magnet:']")
                name_el = it.select_one('span.name')
                if isinstance(name_el, Tag):
                    if no_uncensored and ('-uc' in name_el.text.lower() or '无码' in name_el.text.lower() or '-u.' in name_el.text.lower()):
                        continue
                    if no_subtitle and '字幕' in name_el.text.lower():
                        continue
                
                has_subtitle = False
                tags_el = it.select('.tags span')
                for tag_el in tags_el:
                    if isinstance(tag_el, Tag):
                        if '字幕' in tag_el.text.strip():
                            has_subtitle = True
                            break
                if no_subtitle and has_subtitle:
                    continue
                
                meta_el = it.select_one(".meta")
                if not a or not meta_el:
                    continue
                href_str = a.get("href", "")
                if isinstance(href_str, str):
                    magnet = href_str.strip()
                meta = meta_el.get_text(strip=True)
                sz = size_from_meta(meta)
                if sz > best_size_gib:
                    best_size_gib = sz
                    best_magnet = str(magnet)
        
            # 尝试在评论区找
            if not best_magnet:
                review_html = self._get_soup(f"{target_url}/reviews/lastest")
                if review_html:
                    for content in review_html.select(".review-items .review-item .content"):
                        if content:
                            if 'magnet:' in content.get_text():
                                match = re.search(r"(magnet:\?xt=urn:[a-z0-9]+:[a-zA-Z0-9]{32,})", content.get_text())
                                if match:
                                    best_magnet = str(match.group(1))
                                    break
                            elif 'ed2k:' in content.get_text():
                                match = re.search(r"(ed2k://\|file\|[^ ]+)", content.get_text())
                                if match:
                                    best_magnet = str(match.group(1))
                                    break
        
            result['magnet'] = best_magnet
        
        return result
        
        


import os
from unittest.mock import Mock

import pytest

from loki.scrapers.pan_115 import Pan115, PathExpect


def test_walk_path():
    pan = Pan115.__new__(Pan115)
    pan.path_to_id = Mock(return_value="0")

    def _list_files(dir_id: str, limit: str = "20"):
        if dir_id == "0":
            return [
                {"name": "dirA", "id": "1", "size": 0, "class": "", "is_dir": True},
                {"name": "file1.txt", "id": "f1", "size": 123, "class": "1", "is_dir": False},
            ]
        if dir_id == "1":
            return [
                {"name": "file2.mp4", "id": "f2", "size": 456, "class": "2", "is_dir": False},
                {"name": "dirB", "id": "2", "size": 0, "class": "", "is_dir": True},
            ]
        if dir_id == "2":
            return []
        raise AssertionError(f"unexpected dir_id: {dir_id}")

    pan.list_files = Mock(side_effect=_list_files)

    res = pan.walk_path("/base", limit="20", sleep_seconds=0)
    assert {x["path"] for x in res} == {
        "/base/dirA",
        "/base/file1.txt",
        "/base/dirA/file2.mp4",
        "/base/dirA/dirB",
    }
    by_path = {x["path"]: x for x in res}
    assert by_path["/base/file1.txt"]["id"] == "f1"
    assert by_path["/base/dirA"]["id"] == "1"


def test_list_files_dir_with_extension():
    pan = Pan115.__new__(Pan115)

    resp = Mock()
    resp.json = Mock(return_value={
        "state": True,
        "cid": "0",
        "data": [
            # Folder whose name looks like a file; API may include "class" but has no "fid".
            {"n": "abc.mp4", "cid": "10", "class": "0"},
            {"n": "real.mp4", "fid": "f1", "cid": "0", "s": "123", "class": "1"},
        ],
    })
    pan.request = Mock(return_value=resp)

    files = pan.list_files("0", limit="20")
    dir_item = next(x for x in files if x["name"] == "abc.mp4" and x["is_dir"] is True)
    file_item = next(x for x in files if x["name"] == "real.mp4" and x["is_dir"] is False)
    assert dir_item["id"] == "10"
    assert dir_item["class"] == ""
    assert file_item["id"] == "f1"


def test_path_to_id_expect_type():
    pan = Pan115.__new__(Pan115)

    def _list_files(dir_id: str, limit: str = "20"):
        if dir_id == "0":
            return [
                {"name": "abc.mp4", "id": "10", "is_dir": True},
                {"name": "real.mp4", "id": "f1", "is_dir": False},
                {"name": "dir", "id": "1", "is_dir": True},
            ]
        if dir_id == "1":
            return [{"name": "child", "id": "f2", "is_dir": False}]
        return []

    pan.list_files = Mock(side_effect=_list_files)

    assert pan.path_to_id("/", expect=PathExpect.DIR) == "0"
    assert pan.path_to_id("abc.mp4", expect=PathExpect.DIR) == "10"
    assert pan.path_to_id("abc.mp4", expect=PathExpect.FILE) == ""
    assert pan.path_to_id("real.mp4", expect=PathExpect.FILE) == "f1"
    assert pan.path_to_id("real.mp4", expect=PathExpect.DIR) == ""
    assert pan.path_to_id("dir/child", expect=PathExpect.FILE) == "f2"
    assert pan.path_to_id("real.mp4/child") == ""


def test_paths_to_ids_expect_type():
    pan = Pan115.__new__(Pan115)

    def _list_files(dir_id: str, limit: str = "20"):
        if dir_id == "0":
            return [
                {"name": "abc.mp4", "id": "10", "is_dir": True},
                {"name": "real.mp4", "id": "f1", "is_dir": False},
            ]
        return []

    pan.list_files = Mock(side_effect=_list_files)

    assert pan.paths_to_ids(["abc.mp4", "real.mp4", "missing"], expect=PathExpect.ANY) == ["10", "f1", "-1"]
    assert pan.paths_to_ids(["abc.mp4", "real.mp4"], expect=PathExpect.DIR) == ["10", "-1"]
    assert pan.paths_to_ids(["abc.mp4", "real.mp4"], expect=PathExpect.FILE) == ["-1", "f1"]


def test_delete_passes_expect_and_skips_missing():
    pan = Pan115.__new__(Pan115)
    pan.paths_to_ids = Mock(return_value=["1", "-1", "2"])

    resp = Mock()
    resp.json = Mock(return_value={"state": True})
    pan.request = Mock(return_value=resp)

    ok = pan.delete(["/a", "/missing", "/b"], expect=PathExpect.FILE)
    assert ok is True
    pan.paths_to_ids.assert_called_once()
    assert pan.paths_to_ids.call_args.kwargs["expect"] == PathExpect.FILE
    assert "fid[0]" in pan.request.call_args.kwargs["data"]


def test_move_passes_expect_filters_ids():
    pan = Pan115.__new__(Pan115)
    pan.paths_to_ids = Mock(return_value=["1", "-1", "2"])
    pan.create_folder = Mock(return_value="dest")

    resp = Mock()
    resp.json = Mock(return_value={"state": True})
    pan.request = Mock(return_value=resp)

    ok = pan.move(["/a", "/missing", "/b"], dest_path="/dest", expect=PathExpect.ANY)
    assert ok is True
    data = pan.request.call_args.kwargs["data"]
    assert data["pid"] == "dest"
    assert data["fid[0]"] == "1"
    assert data["fid[1]"] == "2"
    assert "fid[2]" not in data


def test_rename_passes_expect_filters_ids():
    pan = Pan115.__new__(Pan115)
    pan.paths_to_ids = Mock(return_value=["-1", "2"])

    resp = Mock()
    resp.json = Mock(return_value={"state": True})
    pan.request = Mock(return_value=resp)

    ok = pan.rename({"/missing": "a", "/ok": "b"}, expect=PathExpect.FILE)
    assert ok is True
    data = pan.request.call_args.kwargs["data"]
    assert "files_new_name[-1]" not in data
    assert data["files_new_name[2]"] == "b"


def test_add_lixian_task_to_id():
    pan = Pan115.__new__(Pan115)
    pan.uid = "u1"
    pan._offline_space = Mock(return_value={"sign": "s", "time": "t"})

    resp = Mock()
    resp.json = Mock(return_value={"errno": 0, "info_hash": "h", "url": "x"})
    pan.request = Mock(return_value=resp)

    res = pan.add_lixian_task_id(["magnet:?xt=urn:btih:xx"], "123")
    assert res["info_hash"] == "h"
    assert pan.request.call_args.kwargs["data"]["wp_path_id"] == "123"


def test_move_ids_filters_missing():
    pan = Pan115.__new__(Pan115)

    resp = Mock()
    resp.json = Mock(return_value={"state": True})
    pan.request = Mock(return_value=resp)

    ok = pan.move_ids(["1", "-1", "2", ""], "dest")
    assert ok is True
    data = pan.request.call_args.kwargs["data"]
    assert data["pid"] == "dest"
    assert data["fid[0]"] == "1"
    assert data["fid[1]"] == "2"
    assert "fid[2]" not in data


def test_delete_ids_filters_missing():
    pan = Pan115.__new__(Pan115)

    resp = Mock()
    resp.json = Mock(return_value={"state": True})
    pan.request = Mock(return_value=resp)

    ok = pan.delete_ids(["-1", "1", "", "2"])
    assert ok is True
    data = pan.request.call_args.kwargs["data"]
    assert data["fid[0]"] == "1"
    assert data["fid[1]"] == "2"
    assert "fid[2]" not in data


def test_rename_ids_filters_missing():
    pan = Pan115.__new__(Pan115)

    resp = Mock()
    resp.json = Mock(return_value={"state": True})
    pan.request = Mock(return_value=resp)

    ok = pan.rename_ids({"-1": "a", "": "b", "2": "c"})
    assert ok is True
    data = pan.request.call_args.kwargs["data"]
    assert "files_new_name[-1]" not in data
    assert data["files_new_name[2]"] == "c"


def test_create_folder_to_id_creates_missing_segments():
    pan = Pan115.__new__(Pan115)
    pan.create_folder_in = Mock(side_effect=["11"])

    def _list_files(dir_id: str, limit: str = "1000"):
        if dir_id == "0":
            return [{"name": "a", "id": "10", "is_dir": True}]
        if dir_id == "10":
            return []
        raise AssertionError(f"unexpected dir_id: {dir_id}")

    pan.list_files = Mock(side_effect=_list_files)

    out = pan.create_folder_to_id("0", "/a/b", limit="1000")
    assert out == "11"
    pan.create_folder_in.assert_called_once_with("10", "b")


@pytest.mark.skipif(
    os.environ.get("LOKI_PAN115_INTEGRATION") != "1",
    reason="integration test (requires 115 cookies/config and network)",
)
def test_pan115_integration_smoke():
    pan_115 = Pan115(profile_name="acc1")
    res = pan_115.list_files("0")
    assert isinstance(res, list)



from __future__ import annotations

import logging
import time
from enum import Enum
from typing import Any, Mapping

import requests

from loki.common.profile import Profile

logger = logging.getLogger("loki")


class Method(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"


class RequestMixin:
    def __init__(
        self,
        *,
        session: requests.Session | None = None,
        headers: Mapping[str, str] | None = None,
        profile: Profile | None = None,
        auto_load_cookies: bool = True,
    ) -> None:
        self.session = session or requests.Session()
        self.profile = profile
        self.headers: dict[str, str] = dict(headers or {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36'
        })
        if self.profile and auto_load_cookies:
            self.profile.load_cookies_into(self.session)

    def request(
        self,
        method: Method | str,
        url: str,
        *,
        params: Mapping[str, Any] | None = None,
        headers: Mapping[str, str] | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, Any] | None = None,
        max_retries: int = 3,
        backoff_factor: float = 0.5,
        timeout: float = 10,
    ) -> requests.Response | None:
        method_str = method.value if isinstance(method, Method) else str(method)
        merged_headers = {**self.headers, **(dict(headers) if headers else {})}

        for attempt in range(1, max_retries + 1):
            try:
                resp = self.session.request(
                    method_str,
                    url,
                    params=dict(params) if params else None,
                    headers=merged_headers or None,
                    data=data,
                    json=json,
                    files=dict(files) if files else None,
                    timeout=timeout,
                )

                if resp.status_code < 500:
                    return resp

                logger.warning(
                    "Request %s failed with status %s, attempt %s/%s",
                    url,
                    resp.status_code,
                    attempt,
                    max_retries,
                )
            except (requests.Timeout, requests.ConnectionError) as err:
                logger.warning(
                    "Request %s exception %s, attempt %s/%s",
                    url,
                    err,
                    attempt,
                    max_retries,
                )
            except Exception as err:
                logger.exception("Request %s fatal error: %s", url, err)
                return None

            if attempt < max_retries:
                time.sleep(backoff_factor * (2 ** (attempt - 1)))

        return None

    def save_cookies(self) -> None:
        if self.profile:
            self.profile.save_cookies_from(self.session)

    def close(self, *, save_cookies: bool = True) -> None:
        if save_cookies:
            self.save_cookies()
        self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close(save_cookies=True)
        return False

    # Backwards compat with earlier scrapers naming.
    _request = request
